/*
 *  chardev.c: Creates a read-only char device that says how many times
 *  you've read from the dev file
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <asm/uaccess.h>        /* for put_user */
#include <asm/segment.h>
#include <linux/buffer_head.h>
#include <linux/random.h>


/*  
 *  Prototypes - this would normally go in a .h file
 */
int init_module(void);
void cleanup_module(void);
static int device_open(struct inode *, struct file *);
static int device_release(struct inode *, struct file *);
static ssize_t device_read(struct file *, char *, size_t, loff_t *);
static ssize_t device_write(struct file *, const char *, size_t, loff_t *);

#define SUCCESS 0
#define DEVICE_NAME "fruitdev"	/* Dev name as it appears in /proc/devices   */
#define BUF_LEN 40		/* Max length of the message from the device */

/* 
 * Global variables are declared as static, so are global within the file. 
 */

static int Major;		/* Major number assigned to our device driver */
static int Device_Open = 0;	/* Is device open?  
				 * Used to prevent multiple access to device */
static char msg[BUF_LEN] = {'a','a','a','\n'};  /* The msg the device will give when asked */
static char fruit[BUF_LEN] = {'a','a','a','\n'};	/* The msg the device will give when asked */
static char *msg_Ptr;

static struct file_operations fops = {
	.read = device_read,
	.write = device_write,
	.open = device_open,
	.release = device_release
};

/*
 * This function is called when the module is loaded
 */
int init_module(void)
{
        Major = register_chrdev(0, DEVICE_NAME, &fops);

	if (Major < 0) {
	  printk(KERN_ALERT "Registering char device failed with %d\n", Major);
	  return Major;
	}

	printk(KERN_INFO "I was assigned major number %d. To talk to\n", Major);
	printk(KERN_INFO "the driver, create a dev file with\n");
	printk(KERN_INFO "'mknod /dev/%s c %d 0'.\n", DEVICE_NAME, Major);
	printk(KERN_INFO "Try various minor numbers. Try to cat and echo to\n");
	printk(KERN_INFO "the device file.\n");
	printk(KERN_INFO "Remove the device file and module when done.\n");

	return SUCCESS;
}

/*
 * This function is called when the module is unloaded
 */
void cleanup_module(void)
{
	/* 
	 * Unregister the device 
	 */
	unregister_chrdev(Major, DEVICE_NAME);
}

/*
 * Methods
 */

/* 
 * Called when a process tries to open the device file, like
 * "cat /dev/mycharfile"
 */
static int device_open(struct inode *inode, struct file *file)
{
	if (Device_Open)
		return -EBUSY;

	Device_Open++;
        msg_Ptr = msg;
	try_module_get(THIS_MODULE);

	return SUCCESS;
}

/* 
 * Called when a process closes the device file.
 */
static int device_release(struct inode *inode, struct file *file)
{
	Device_Open--;		/* We're now ready for our next caller */

	/* 
	 * Decrement the usage count, or else once you opened the file, you'll
	 * never get get rid of the module. 
	 */
	module_put(THIS_MODULE);

	return 0;
}

/* 
 * Called when a process, which already opened the dev file, attempts to
 * read from it.
 */
static ssize_t device_read(struct file *filp,	/* see include/linux/fs.h   */
			   char *buffer,	/* buffer to fill with data */
			   size_t length,	/* length of the buffer     */
			   loff_t * offset)
{
	/*
	 * Number of bytes actually written to the buffer 
	 */
	int bytes_read = 0;

	/*
	 * If we're at the end of the message, 
	 * return 0 signifying end of file 
	 */
	if (*msg_Ptr == 0)
		return 0;

	/* 
	 * Actually put the data into the buffer 
	 */
	while (length && *msg_Ptr) {

		/* 
		 * The buffer is in the user data segment, not the kernel 
		 * segment so "*" assignment won't work.  We have to use 
		 * put_user which copies data from the kernel data segment to
		 * the user data segment. 
		 */
		put_user(*(msg_Ptr++), buffer++);

		length--;
		bytes_read++;
	}

	/* 
	 * Most read functions return the number of bytes put into the buffer
	 */
	return bytes_read;
}






static int read_line(struct file * f, char * buf, unsigned int line, size_t max_len) {
            mm_segment_t oldfs;
            unsigned long long offset = 0;
            char b[10];
            char * p = b;
            const char * p_end = &b[9];
            char * bufp = buf;
            size_t len = 0;
            unsigned int current_line = 1;
            int read_length;
            
            if (line < 1)
                return -EINVAL;
            
            // Get current segment descriptor
            oldfs = get_fs();
            // Set segment descriptor associated to kernel space
            set_fs(get_ds());
            // Read the file
            vfs_read(f, b, 9, &offset);
            while (1) {
                // are there chars left?
                if (p == p_end) {
                    p = b;
                    read_length = vfs_read(f, b, 9, &offset);
                    if (read_length <= 0)  // no more chars left, no newline found
                        return -1;
                }
                
                // is there a new line
                if (*p == '\n') {
                    if (current_line++ == line) {
                        *bufp = '\0';
                        return len;
                    }
                }
                else if (line == current_line) {
                    *bufp = *p;
                    bufp++;
                    len++;
                    // line too long
                    if (len > max_len)
                        return -1;
                }
                p++;
            }
            // Restore segment descriptor
            set_fs(oldfs);
            
            return 0;
}

static void file_close(struct file* file) {
            filp_close(file, NULL);
}


static struct file* file_open(const char * path, int flags) {
            struct file * filp;
            mm_segment_t oldfs;

            oldfs = get_fs();
            set_fs(get_ds());
            filp = filp_open(path, flags, S_IRWXU);
            set_fs(oldfs);
            if(IS_ERR(filp)) {
                return NULL;
            }
            return filp;
}

static void shuffle_letters(char * buf, size_t len) {
            size_t i;
            unsigned int r;
            char t;
            for (i = 0; i < len; i++) {
                get_random_bytes(&r, sizeof(r));
                r = r % len;
                
                t = buf[r];
                buf[r] = buf[i];
                buf[i] = t;
            }
}

static void add_line(char * buf, size_t len) {
    
}

/*  
 * Called when a process writes to dev file: echo "hi" > /dev/hello 
 */
static ssize_t device_write(struct file *filp, const char *buff, size_t len, loff_t * off)
{
        static unsigned int line = 1;
        ssize_t line_length;
        
	if (len <= 0)
            return -EINVAL;
        
        if ( strncmp(buff, "next", 4) == 0 ) {
            // read next fruit, mix it, increment line number
            struct file * f;
            
            f = file_open("fruits.txt", O_RDONLY);
            if (!f)
                return -EBADFD;
            
            line_length = read_line(f, fruit, line++, BUF_LEN);
            if (line_length <= 0) {
                file_close(f);
                
                // retry once, in case of too high line
                line = 1;
                f = file_open("fruits.txt", O_RDONLY);
                if (!f)
                    return -EBADFD;
                line_length = read_line(f, fruit, line++, BUF_LEN);
                if (line_length <= 0) {
                    file_close(f);
                    return -EBADF;
                }
            }
            
            file_close(f);
            
            memcpy(msg, fruit, line_length+1);
            shuffle_letters(msg, strlen(msg));
            
        }
        else if ( strncmp(buff, "solution", 8) == 0 )
            memcpy(msg, fruit, strlen(fruit)+1);
        else if ( strncmp(buff, "repeat", 6) == 0 ) {
            memcpy(msg, fruit, strlen(fruit)+1);
            shuffle_letters(msg, strlen(msg));
        }
        else if ( strncmp(buff, "answer", 6) == 0 ) {
            if (strncmp( buff + 7, fruit, strlen(fruit)))
                strcpy(msg, "No, try again!");
            else
                strcpy(msg, "Yeah!");
        }
        else if ( strncmp(buff, "add", 3) == 0 ) {
            char * pc;
            struct file * f;
            
            f = file_open("fruits.txt", O_WRONLY);
            if (!f)
                return -EBADFD;
            
            pc = strchr(buff, '\n');
//             msg[0] = pc - buff + 0x30;
            
            
        }

        return len;
}
